Attribute VB_Name = "MainStuff"
'
'  Perceptra-1D, Perception Software using a tree of Binons (Binary Neurons).

'  Started 28th July 2012  (c) Copyright Adaptron Inc.
'  This Version was created 16th October 2017 for general distribution to the public
'
'   This program learns to recognize handwriiten digits from 0 to 9.
'   The input is a list of 5607 handwritten digits stored in the "Hand Drawn Digits 8x8.txt" file.
'   Each line in the file contains 65 numbers separated by commas.
'   The first 64 numbers are the 8 by 8 pixel intensities for the image of the digit.
'   The pixel intensity values in the file range from 0 for white to 16 for black.
'   The software adds one to the pixel values so that the zeros become ones because the log(0) is not valid.
'   Therefore the sesnor readings range from 1 (white) to 17 (black) when processed for recognition.
'   The 65th number is the actual digit label from 0 to 9.
'   Perceptra starts with no binons before recognizing the images
'

  DefInt A-Z            '16 bits / 2 bytes
  
          'ProcessNextDigit_Click() on the MainDisplay calls RecognizeDigit() and then MakePrediction()
          'All the binons found in the image are kept in the ActiveBinonTree() and the AllFoundBinons() list.
          'A subset of the AllFoundBinons() that are associated with only one label binon are put in the FoundBinons()
          'for determining the predicted digit label.
          'After it has made a prediction then it updates the binons with the actual digits as associated Label binons
          'RecognizeDigit() calls DoCenterSurround() to convert the sensor readings into level 0 binons based on finding the
          'edges. The center surround is an edge detection technique used by the human eye.
          'Then it calls CountRepeaters() and CombineBinons() repeatedly at each level of complexity
          'to create more complex shape patterns
          'The ActiveBinonTree() array is used to keep the Shape Binons for combining at higher levels.
          'ActiveBinonTree() is hierarchical and the process does breadth first binon creation.
  
  Public DigitPixels(1 To 6000, 7, 7)   'This is an 8 by 8 array of pixels that have values 0 thru 16
                                        'Each 8 by 8 array is an image of a handwritten digit.
  Public ActualDigit(1 To 6000)   'the label for the actual digit in the image.
  Public Number_of_Digits         'The number of digits in DigitPixels()
  Public Stimulus_Number          'Index into DigitPixels() and ActualDigit() arrays for the image being recognized
                                  'It is increased by 1 every time a digit image is processed
                                  
  Public BitmapsOrImages$         'Processing black and white bitmaps or pixel images of the digits
  Public Bitmaps$                 '="Bitmaps"     'A pixel value 7 or less becomes white, A pixel value > 7 becomes black
  Public Images$                  '="Images"      'DoCenterSurround() determines the sensor values to use
  
  Public Processing50 As Boolean  'Flag set so that if doing 50, no reporting to display until end of 50 digits processed
  
'--------------------------  Variables used to determine the best prediction for the digit's label
  Public FindCountForDigit(0 To 9)    'The number of unambiguous shape binons found for each digit label from 0 to 9
  Public HighestFindLevel(0 To 9)     'The highest binon level (most complex pattern) that was found for each digit label in the digit image
                                  
  Public AllFoundBinons&(0 To 500)    'All the binons found during the processing of one image,
                                      'AllFoundBinons(0) is the count of entries in the array
  Public FoundBinons&(0 To 100, 2)    'Up to 100 binons found with FoundBinons(0, 0) = count of entries
                '---- FoundBinons() properties - its 2nd index
  Public TheFoundBinon            ' 0 = Found Binon ID
  Public TheLabelBinon            ' 1 = Label Binon ID
  Public TheBinonLevel            ' 2 = Binon level
                                  
'--------------------------  Variables to keep track of the prediction accuracy in the last 100 digits processed ---------
  Public IsACorrectPrediction(1 To 100) As Boolean  'Flagged True if it was a correct prediction for the last 100 digits
  Public NextEntryIn100             'location in IsACorrectPrediction() to put the next one
  Public NumberOfEntriesInLast_100  'number of entries in IsACorrectPrediction()
 
  Public PredictedDigitLabel        'The digit prediction = -1 if no prediction made
    
  Public CorrectInLast50            'The count of digits predicted correctly in the last 50 digits
  Public NoPredictionMade           'The count of when no prediction made
  
  Public DigitCount(0 To 9)         'Number of times this digit experienced
  Public CorrectDigit(0 To 9)       'Number of times this digit predicted correctly
  
  Public LvlCorrect(0 To 12)        'Number of correct predictions at this level
  Public LvlPredictions(0 To 12)    'Number of predictions at this level, 12 = 12 or more
  
' --------------------------  Sensors --------------------------

  Public NumberOfSensors          'The number of sensors = 64
  Public SensorReadings(1 To 64)  'The sensor readings (one based)
  
  '------------------  The Binons  = Long Term Memory -------------------------
  'This is a list of all the Binons. Each one has 7 properties
                                
  Public Binons&(1 To 120000, 1 To 7)        'The Binons
                                '1st index = the binon number/ID
                                '2nd index = binon property
  Public NumberOfBinons&        'Number of binons so far in Binons() for all the binon types, incremented before it is used
                                'NumberOfBinons + 1 = Place where next new binon will be placed in Binons()
  Public MaxBinons&             '=120000, Maximum number of binons
  
  Public BinonTypes             '3, number of binon types, includes shapes, labels and associations
                                '1=ShapeBin, 2=LabelBin, 3=AssociationBin
  
                '---- Binon Properties - 2nd index values
  Public BinonType              '1, The type of binon (ShapeBinon, LabelBinon, AssociationBinon)
  Public TriggerSource          '2, Pointer to the Trigger source binon
  Public GoalSource             '3, Pointer to the Goal source binon
  Public Ratio                  '4, (Integer Difference of Logs - IDL) ratio value (based on JND - Just Noticeable Diiference)
  Public Interest               '5, Interest Value - see below
  Public BinonLevel             '6, Level of binon
  Public AssociationList        '7, link to AssociationLinks - list of Associated label binons, = NullLink if empty
  
  Public NullBinon&             '-1, Indicates the TriggerSource or GoalSource pointer is not pointing to another binon
                                'Used in level 0 binons (sensor level)
  
                '---- Interest values for Binons(x, Interest)
  Public NOVEL              '2  New - 1st time experienced
  Public FAMILIAR           '1  Neutral, experienced 2 or more times
  Public NoInterest         '-1 Interest not known yet
  
                '---- The values for binon types
                                '1=ShapeBin, 2=LabelBin, 3=AssociationBin
  Public ShapeBin           '1, = Shapes = patterns of black and white
  Public LabelBin           '2, = Symbolic binon representing the label (digits 0 thru 9 - value kept in the Ratio property)
  Public AssociationBin     '3, = Association binon between ShapeBin & LabelBin
  
  Public NullBinonType      '0, Indicates the type not set, binon will be NullBinon
  
  Public AssociationLinks&(1 To 120000, 1 To 2)  'Linked list of AssociationBinons,  - set up in Reset_Adaptron
                  '---- AssociationLinks() 2nd index property values
  Public NextLink           '1 = The pointer to the next AssociationLinks() entry in this linked list
                            '    it may be a NullLink at the end of the linked list
  Public AssocBinonID       '2 = an AssociationBinon ID
  
  Public NullLink&          '=0, A NextLink pointer at the end of the linked list
    
  Public FreeLink&          'Pointer to start of linked list of free AssociationLinks
  
  '---------------------  Activation Tree  ------------------------------------------
  
  Public ActiveBinonTree&(0 To 25, 1 To 64, 1 To 3) 'The tree of the Shape Binons found in the image.
                                        'Dynamically recreated as each image is processed/experienced/perceived
                                        '1st index = which level is being processed
                                        '2nd index = The active entry - corresponds to a sensor position
                                        '3rd index = Activation property values
  
                '---- ActiveBinonTree() types of property values, 3rd index
  Public PatternSize    '1, The size of the pattern - measured in number of sensors involved at level 0
  Public Quantity       '2, Numerosity, quantity of times this pattern repeats - zero if entry is to be skipped/empty
  Public TheBinon       '3, The binon / pattern experienced - if any - a ShapeBin
  
  Public ActiveLevel                    'Which of the ActiveBinonTree levels is being processed
  Public HighestActiveLevel             'The highest level of active binons after an image has been processed
  Public MaxLevels                      '=25, The highest level of complexity to which the active binon tree can grow
  
  '---------------------  Other Variables ------------------------------------------
    
  Public OverLimit As Boolean           'Global flag set when over Binon list limit
  Public CountCombined                  'Count of binon combinations performed as each level is processed
                                        'If it is one or less then stop the recognition process for the image
  
  Public LogBase#         '1.2 for a 20% just noticeable difference
                          'If it is 2.0 then that is a 100% just noticeable difference
  
  '---------------------- Operational Values - Adaptron --------------------
  
  Public ParallelSign$    '" + "
 
  Public I, J, T
  
Public Sub Main()                 'Program Starts Here
          
  Bitmaps$ = "Bitmaps"            'Black and white images
  Images$ = "Images"              'Pixel value / gray scale images
  
  NumberOfSensors = 64            'The number / range of sensors in the Sense
                          
                '---- FoundBinons() properties - its 2nd index
  TheFoundBinon = 0       ' Found Binon ID
  TheLabelBinon = 1       ' Label Binon ID
  TheBinonLevel = 2       ' Binon level
  
  '------------------  The Long Term Memory (Binons)  -------------------------
  
  MaxBinons = 120000      'Max number of binons - all types stored in the one array
  
  BinonTypes = 3          'number of binon types, includes label names and associations
                          '1=ShapeBin, 2=LabelBin, 3=AssociationBin
  
                '---- Binons() Properties - 2nd index values
  BinonType = 1           ' The type of binon (ShapeBinon, LabelBinon, or AssociationBinon)
  TriggerSource = 2       ' Pointer to the Trigger source binon
  GoalSource = 3          ' Pointer to the Goal source binon
  Ratio = 4               ' (Integer Difference of Logs - IDL) ratio value
  Interest = 5            ' Interest Value - see below
  BinonLevel = 6          ' Level of binon
  AssociationList = 7     ' Link to AssociationLinks - list of Associated label binons, = NullLink if empty
  
                '---- The values for stimulus binon types         '1=ShapeBin, 2=LabelBin, 3=AssociationBin
  ShapeBin = 1            ' Shapes
  LabelBin = 2            ' Symbolic binon representing the label
  AssociationBin = 3      ' Association binon between ShapeBin & LabelBin
  
  NullBinon = -1          ' Indicates the TriggerSource or GoalSource pointer is not pointing to another binon
  NullBinonType = 0       ' Indicates the type not set, binon will be NullBinon
  
                '---- Interest values for Binons(x, Interest)
  NOVEL = 2               '  New - 1st time experienced
  FAMILIAR = 1            '  Neutral, experienced 2 or more times
  NoInterest = -1         '  Interest not known yet
   
                '---- AssociationLinks() 2nd index values
  NextLink = 1            'The pointer to the next AssociationLinks() entry in this linked list
  AssocBinonID = 2        'an AssociationBinon ID
  NullLink = 0            'A NextLink pointer at the end of the linked list
  
                '---- ActiveBinonTree() types of information values, 3rd index
  PatternSize = 1         ' The size of the pattern - measured in number of sensors involved at level 0
  Quantity = 2            ' Numerosity, quantity of times this pattern repeats - zero if entry is to be skipped
  TheBinon = 3            ' The binon / pattern experienced ShapeBin - if any
  
  HighestActiveLevel = 0  ' The highest level of active binons after a image has been processed
  MaxLevels = 25          ' The highest level of complexity to which the binon tree can grow
  
  '---------------------- Operational Values -------------------------------
  
  ParallelSign$ = " + "     '" + "
  
'         'Start by loading the 8x8 images of the digits
  Call Get8x8Digits                 'Load the data into DigitPixels()
  
  MainDisplay.Show                  'Will Load the form and display it and set up initial settings
  
End Sub             'End of Main()

Public Sub GetTypeInfo(BinTyp, BinonTyp$, AbbreviatedType$)

                                '1=ShapeBin, 2=LabelBin, 3=AssociationBin
    BinonTyp$ = ""
    AbbreviatedType$ = ""
        
    Select Case BinTyp
        
      Case ShapeBin               'Shape
        BinonTyp$ = "Shapes"
        AbbreviatedType$ = "Shp= "
                
      Case LabelBin               'Digit Label, 0 to 9
        BinonTyp$ = "Labels"
        AbbreviatedType$ = "Lab= "
        
      Case AssociationBin         'Association between a shape and a digit label
        BinonTyp$ = "Label Associations"
        AbbreviatedType$ = "Ass= "
                            
      Case NullBinonType
        BinonTyp$ = "Null"
        AbbreviatedType$ = "~  = "
        
      Case Else
        BinonTyp$ = "Unknown"
        AbbreviatedType$ = "XXX= "
        
    End Select
                
End Sub           'End of GetTypeInfo()

Public Function TypeOfBinon(TheBinonID&)       'Return the type of the binon

    If TheBinonID = NullBinon Then
      TypeOfBinon = NullBinonType
    Else
      TypeOfBinon = Binons(TheBinonID, BinonType)
    End If

End Function      'End of TypeOfBinon()

Public Sub GetValue(BinonID&, ReturnValue$)

              'Recursive routine to add values to ReturnValue$ for any binon/pattern
              'Returns the value for the binon pattern based on its binon type
              
  Dim BinonIDTemp&, Reading
              
    If BinonID = NullBinon Then Exit Sub
    
    Select Case TypeOfBinon(BinonID)
    
'  Public ShapeBin           '1, = Shape binon
      Case ShapeBin
        BinonIDTemp = BinonID
        If Binons(BinonIDTemp, TriggerSource) <> NullBinon Then   'not at sensor level 0
          While Binons(BinonIDTemp, BinonLevel) > 1               'go all way down left side to level-1
            BinonIDTemp = Binons(BinonIDTemp, TriggerSource)      'this is the ShapeBin
          Wend
        End If
                            'BinonIDTemp must have be at level-1 so get the value of the Ratio
        Reading = Binons(BinonIDTemp, Ratio)                      'Value is in Ratio at level 1 for readings
        ReturnValue$ = ReturnValue$ & Trim(Str$(Reading))
        
        If Binons(BinonID, BinonLevel) > 1 Then
          ReturnValue$ = ReturnValue$ & ParallelSign$
          Call GetValue(Binons(BinonID, GoalSource), ReturnValue$) 'Do the right side - will return if = NullBinon
        End If
        
'  Public LabelBin           '3, = Symbolic binon representing the label of the image (the actual digit)
      Case LabelBin
        Reading = Binons(BinonID, Ratio)                          'Label value is at sensor Binon Level=0
        ReturnValue$ = ReturnValue$ & Trim(Str$(Reading))
      
      ' does not do Association binons
      
    End Select

End Sub       'End of GetValue()

Public Sub IdBinons(BinLevel, BinType, Bin1&, Bin2&, IDLRatio, BinonID&)  'Create new or find existing binon

  'BinLevel = Level of the binon, 0 primitive values - sensor level, 1+ patterns with ratio values
  'BinType  = Binon Type
  'Bin1 = The trigger source binon
  'Bin2 = The goal source binon
  'IDLRatio = The Ratio value (Integer Difference of Logs - IDL)
  'BinonID  = The pattern / binon identifier returned, = NullBinon if over the limit
  
            'Given two binons Bin1 (Trigger) and Bin2 (Goal) and an IDL Ratio
            'see if we have a binon with the same values and the same Bin1 and Bin2.
            'Return the found or newly created binon
  
'  Public BinonType              '1, The type of binon (ShapeBinon, LabelBinon, AssociationBinon)
'  Public TriggerSource          '2, Pointer to the Trigger source binon
'  Public GoalSource             '3, Pointer to the Goal source binon
'  Public Ratio                  '4, (Integer Difference of Logs - IDL) ratio value
'  Public Interest               '5, Interest Value - see below
'  Public BinonLevel             '6, Level of binon
'  Public AssociationList        '7, link to AssociationLinks - list of Associated label binons, = NullLink if empty

    BinonID = NullBinon                                 'Return no binon - none created by default
    
                    'First determine if we have any binons of the given type
    If NumberOfBinons > 0 Then      'if none then do not check if ExistBinon
                    'If the binons already exists then
        'ExistBinon(BinLevel, BinType, Bin1&, Bin2&, RatioValue, BinonID&)  will return the one found in BinonID
      If ExistBinon(BinLevel, BinType, Bin1, Bin2, IDLRatio, BinonID) Then
        Binons(BinonID, Interest) = FAMILIAR            'Novel -> Familiar or it remains familiar
      End If
    End If
    
    If BinonID = NullBinon Then                         'If it does not already exist then create it
    
      If NumberOfBinons >= MaxBinons Then               'if can not add any more binons to memory
        OverLimit = True
        Exit Sub
      End If
    
      BinonID = NumberOfBinons + 1                      'it was not found - so add it as a new one
      
      Binons(BinonID, BinonType) = BinType              'The type of binon
      Binons(BinonID, TriggerSource) = Bin1             'The trigger source binon
      Binons(BinonID, GoalSource) = Bin2                'The goal source binon

      Binons(BinonID, Ratio) = IDLRatio                 'stored binon ratio values
              'Level 0 sensor binons become familiar immediately
      If BinLevel = 0 Then
        Binons(BinonID, Interest) = FAMILIAR            'Familiar - sensor reading level
      Else
        Binons(BinonID, Interest) = NOVEL               'Patterns at level 1+ start out novel
      End If
        
      Binons(BinonID, BinonLevel) = BinLevel            'The complexity level of the binon being created
      Binons(BinonID, AssociationList) = NullLink       'no associated label binon links
      NumberOfBinons = BinonID
      
    End If    'End if not ExistBinon
    
End Sub               'End of IdBinons()

Public Function ExistBinon(BinLevel, BinType, Bin1&, Bin2&, RatioValue, BinonID&) As Boolean
                            
    'Given two binons numbered Bin1 and Bin2 with Ratio of RatioValue see if we have a binon in Binons()
    'that matches this combination of type BinType, if so return its number in BinonID
    'BinonID = NullBinon if not found and ExistBinon is set to False
    
    Dim BinID&
    
    ExistBinon = False                                  'assume it does not exist
    BinonID = NullBinon                                 'return NullBinon if not found
    
    For BinID = 1 To NumberOfBinons
                      'Binon level is an identifying property of a binon
      If Binons(BinID, BinonLevel) = BinLevel Then
        If Binons(BinID, BinonType) = BinType Then      'If it is the same binon type then
                      'TriggerSource = Binon source 1 pointer and GoalSource = Binon source 2 pointer
                      'Given binon IDs - search for same pair of binons
          If Binons(BinID, TriggerSource) = Bin1 And Binons(BinID, GoalSource) = Bin2 Then
            If RatioValue = Binons(BinID, Ratio) Then   'If it has the same (Integer Difference of Logs - IDL) Ratio then
              BinonID = BinID
              ExistBinon = True
              BinID = NumberOfBinons                    'terminate the loop
            End If
          End If
        End If
      End If
    Next BinID

End Function          'End of ExistBinon()

Public Function InterestIn(BinonID&) As Integer

    InterestIn = Binons(BinonID, Interest)
    
End Function          'End of Function InterestIn()

Public Sub ReportBinons()
        
            'Reports on ActiveBinonTree in ActiveList textbox
            'Reports on Binons in BinonList textbox
            
  Dim Pattern$, FirstLine$, MoreLines$, AllLines$
  Dim ThisBinon$, CountBinons, Entry
  Dim ActiveBinon&, BinTyp$, TypeAbbreviated$
  Dim ActiveLvl
  
    If OverLimit Then Exit Sub
  
    MainDisplay.ActiveList.Text = ""
                    
    For ActiveLvl = 0 To HighestActiveLevel                               'for each level that binons were kept
  
        AllLines$ = ""
    
                '---- ActiveBinonTree() types of information values, 3rd index
'  Public PatternSize    '1, The size of the pattern - measured in number of sensors involved at level 0
'  Public Quantity       '2, Numerosity, quantity of times this pattern repeats - zero if entry is to be skipped
'  Public TheBinon       '3, The binon / pattern experienced - if any - ShapeBin

        For Entry = 1 To NumberOfSensors
          ActiveBinon = ActiveBinonTree(ActiveLvl, Entry, TheBinon)       'The combined binon in Activation Tree
          If ActiveBinon <> NullBinon Then                                'If there is a binon here then
            ThisBinon$ = Str$(ActiveBinon)
            Pattern$ = " "
            Call GetValue(ActiveBinon, Pattern$)
          Else                                                            'Else it is an empty ActiveBinonTree entry
            ThisBinon$ = " "
            Pattern$ = ""
          End If
        
          AllLines$ = AllLines$ & Right$("    " & Str$(Entry), 5) & _
                      Right$("   " & Str$(ActiveBinonTree(ActiveLvl, Entry, PatternSize)), 4) & _
                      Right$("   " & Str$(ActiveBinonTree(ActiveLvl, Entry, Quantity)), 5) & _
                      Right$("    " & ThisBinon$, 5) & _
                      " " & Pattern$ & _
                      vbCrLf

        Next Entry
    
        AllLines$ = AllLines$ + vbCrLf
        Call GetTypeInfo(ShapeBin, BinTyp$, TypeAbbreviated$)
        MainDisplay.ActiveList.Text = MainDisplay.ActiveList.Text & "Level=" & Str$(ActiveLvl) & _
                              ", Binon Type=" & BinTyp$ & vbCrLf & AllLines$ & vbCrLf
    Next ActiveLvl
    
  Dim BinonID&, InterestInfo$
  Dim BinType, TrigBinType, GoalBinType, TrigType$, GoalType$, AbbrevTrigType$, AbbrevGoalType$
  
                '---- Binon Properties - 2nd index values
'  Public BinonType              '1, The type of binon (ShapeBinon, LabelBinon, AssociationBinon)
'  Public TriggerSource          '2, Pointer to the Trigger source binon
'  Public GoalSource             '3, Pointer to the Goal source binon
'  Public Ratio                  '4, (Integer Difference of Logs - IDL) ratio value
'  Public Interest               '5, Interest Value - see below
'  Public BinonLevel             '6, Level of binon
'  Public AssociationList        '7, link to AssociationLinks - list of Associated label binons, = NullLink if empty
  
    AllLines$ = ""
    For BinType = 1 To BinonTypes                     '1=ShapeBin, 2=LabelBin, 3=AssociationBin

      MoreLines$ = ""
      Call GetTypeInfo(BinType, BinTyp$, TypeAbbreviated$)
      FirstLine$ = BinType & ", " & BinTyp$
      CountBinons = 0
      
      For BinonID = 1 To NumberOfBinons
      
        If TypeOfBinon(BinonID) = BinType Then        'If it is of this type then
          CountBinons = CountBinons + 1
          Pattern$ = ""
          Call GetValue(BinonID, Pattern$)
          
          TrigBinType = TypeOfBinon(Binons(BinonID, TriggerSource))
          Call GetTypeInfo(TrigBinType, TrigType$, AbbrevTrigType$)
          
          GoalBinType = TypeOfBinon(Binons(BinonID, GoalSource))
          Call GetTypeInfo(GoalBinType, GoalType$, AbbrevGoalType$)
          
          InterestInfo$ = InterestSymbol$(BinonID)
       
          MoreLines$ = MoreLines$ & Right$("    " & Str$(BinonID), 5) & _
                  Right$("    " & Str$(Binons(BinonID, BinonLevel)), 4) & _
                  "  " & Left$(AbbrevTrigType$, 3) & _
                  Right$("    " & Str$(Binons(BinonID, TriggerSource)), 5) & _
                  "  " & Left$(AbbrevGoalType$, 3) & _
                  Right$("    " & Str$(Binons(BinonID, GoalSource)), 5) & _
                  Right$("    " & Str$(Binons(BinonID, Ratio)), 5) & _
                  Right$("    " & InterestInfo$, 5) & _
                  "  " & Pattern$ & _
                  vbCrLf

        End If
        
        If BinType = ShapeBin And CountBinons > 500 Then
          BinonID = NumberOfBinons                     'terminate the loop - only display 1st 500 Shape binons
          MoreLines$ = MoreLines$ & " **** Only first 500 Shape Binons displayed ****" & vbCrLf
        End If
        
        If BinType = AssociationBin And CountBinons > 300 Then
          BinonID = NumberOfBinons                     'terminate loop - only display 1st 300 Association binons
          MoreLines$ = MoreLines$ & " **** Only first 300 Association Binons displayed ****" & vbCrLf
        End If
        
      Next BinonID
      AllLines$ = AllLines$ + FirstLine$ & "  Total Binons=" & Str$(CountBinons) & vbCrLf & MoreLines$ & vbCrLf
      
    Next BinType
    
    MainDisplay.BinonList.Text = AllLines$
  
End Sub             'End ReportBinons()

Public Sub ReportPrediction()

    Dim I, TotalCorrectPreds, TotalPreds, NumberOfPredictionsCorrect, PredictionText$, BinID&
    
    If OverLimit Then Exit Sub
    
      If PredictedDigitLabel <> -1 Then             'The digit prediction = -1 if no prediction made
        MainDisplay.ThePrediction.Text = "Prediction=" & Str$(PredictedDigitLabel)
        
        NumberOfPredictionsCorrect = 0
        For I = 1 To NumberOfEntriesInLast_100      'Count how many were correctly predicted
          If IsACorrectPrediction(I) Then NumberOfPredictionsCorrect = NumberOfPredictionsCorrect + 1
        Next I
 
        NumberOfPredictionsCorrect = NumberOfPredictionsCorrect * 100 / NumberOfEntriesInLast_100  'percent correct
 
        MainDisplay.PercentCorrect.Text = Str$(NumberOfPredictionsCorrect) & " %"

      Else
        MainDisplay.ThePrediction.Text = "No Prediction !"
      End If
      
'    --------------------  Update the Digit correct %s  ----------------
'  Public DigitCount(0 To 9)       'Number of times this digit experienced
'  Public CorrectDigit(0 To 9)     'Number of times this digit predicted Correctly
  
      For I = 0 To 9
        MainDisplay.CountDigits(I).Text = "of" & Str$(DigitCount(I))
        If DigitCount(I) > 0 Then
          MainDisplay.DigitCorrect(I).Text = _
                        Str$(Int(CorrectDigit(I) / DigitCount(I) * 100)) & " %"
        Else
          MainDisplay.DigitCorrect(I).Text = " 0 %"
        End If
      Next I
  
'  Public LvlCorrect(0 To 12)      'Number of correct predictions at this level
'  Public LvlPredictions(0 To 12)  'Number of predictions at this level

      TotalCorrectPreds = 0
      TotalPreds = 0
      For I = 0 To 12
        MainDisplay.Predictions(I).Text = "of" & Str$(LvlPredictions(I))
        If LvlPredictions(I) > 0 Then
          TotalCorrectPreds = TotalCorrectPreds + LvlCorrect(I)
          TotalPreds = TotalPreds + LvlPredictions(I)
          MainDisplay.Corrects(I).Text = Str$(Int(LvlCorrect(I) / LvlPredictions(I) * 100)) & " %"
        Else
        End If
      Next I
      MainDisplay.TotalPredictions.Text = "of" & Str$(TotalPreds)
      MainDisplay.NoPrediction.Text = Str$(NoPredictionMade)
      If TotalPreds > 0 Then
        MainDisplay.CorrectTotal.Text = Str$(Int(TotalCorrectPreds / TotalPreds * 100)) & " %"
      End If
  
'  Public FoundBinons&(0 To 100, 3) 'Up to 100 binons found with (0, 0) = count of entries
                '---- FoundBinons() properties - its 2nd index
'  Public TheFoundBinon            ' 0 = Found Binon ID
'  Public TheLabelBinon            ' 1 = Label Binon ID
'  Public TheBinonLevel            ' 2 = Binon level
                                  
      PredictionText$ = ""
      For I = 1 To FoundBinons(0, 0)
                              'The Binon Number, The binon level, and the predicted digit label
        PredictionText$ = PredictionText$ & _
                    Right$("   " & Str$(FoundBinons(I, TheFoundBinon)), 5) & _
                    "    " & Str$(FoundBinons(I, TheBinonLevel)) & _
                    "    " & Str$(Binons(FoundBinons(I, TheLabelBinon), Ratio)) & vbCrLf
      Next I
      MainDisplay.PredictionList.Text = PredictionText$

    Dim CountEm&, Count12&, L&
    
        MainDisplay.ShapeTotal.Text = "0"

        For L = 0 To 11                                   'For the 1st 11 levels of binons
          CountEm = 0                                     'Count ShapeBin Binons levels 0 to 11
          For BinID = 1 To NumberOfBinons
            If Binons(BinID, BinonLevel) = L Then         'If at this level then
              If TypeOfBinon(BinID) = ShapeBin Then
                CountEm = CountEm + 1
              End If
            End If
          Next BinID
          
          MainDisplay.TotalShapes(L).Text = Str$(CountEm)
          MainDisplay.ShapeTotal.Text = MainDisplay.ShapeTotal.Text + CountEm
        Next L
        
        Count12 = 0                                       'Count for Binons at level 12 or higher
        For BinID = 1 To NumberOfBinons
          If Binons(BinID, BinonLevel) > 11 Then
            If TypeOfBinon(BinID) = ShapeBin Then
              Count12 = Count12 + 1
            End If
          End If
        Next BinID
        
        MainDisplay.TotalShapes(12).Text = Str$(Count12)
        MainDisplay.ShapeTotal.Text = MainDisplay.ShapeTotal.Text + Count12
        
End Sub             'End of ReportPrediction()

Public Function InterestSymbol$(BinonID&)

  InterestSymbol$ = IntrstSymbol$(InterestIn(BinonID))
  
End Function        'End of InterestSymbol$()

Public Function IntrstSymbol$(Intrst)

  IntrstSymbol$ = " "
  
  Select Case Intrst
  
    Case NOVEL
      IntrstSymbol$ = Chr$(254)           'A small black box symbol
      
    Case FAMILIAR
      IntrstSymbol$ = "o"
      
  End Select
  
End Function        'End of IntrstSymbol$()

Public Sub RecognizeDigit(DigitPixels())

        'Called by ProcessNextDigit for each digit image

  Dim SensorReading
  Dim SensorNumber                                          'The pixel position being processed

    For SensorNumber = 1 To NumberOfSensors
      SensorReading = DigitPixels(Stimulus_Number, (SensorNumber - 1) Mod 8, (SensorNumber - 1) \ 8)
      If BitmapsOrImages$ = Bitmaps$ Then
        If SensorReading > 7 Then
          SensorReading = 17                                'black
        Else                                                'or
          SensorReading = 1                                 'white
        End If
      Else                                                  'else gray scale pixel values
        SensorReading = SensorReading + 1                   'Pixel values plus 1 = values from 1 to 17
      End If
      SensorReadings(SensorNumber) = SensorReading          'Sensors are numbered 1 to 64
    Next
    
    If Not Processing50 Then
      Call MainDisplay.DisplayLinearArray(SensorReadings()) 'Sensor values for processing are from 1 to 17
    End If
    
      ActiveLevel = 0
       
      Call ClearActivationTree              'for each digit first clear the activation tree
      
      Call DoCenterSurround(SensorReadings(), ActiveLevel)  'Create edges based on Center Surround On and Offs
                                                            'and populate level 0 with black and white binons
                                            'The patterns will be combinations of level 0 binons
      CountCombined = NumberOfSensors       ' > 0 so the While loop is done at least once
      While CountCombined > 1               'stop combining binons when less than 2 produced at previous level
        CountCombined = 0                   'incremented by CombineBinons() if any patterns found at ActiveLevel+1
                'If a pattern repeats with the adjacent one and is the same size then it is counted as a repeat.
                'It will have a quantity > 1. Remove the repeating ones from The ActiveBinonTree, they have been counted.
        Call CountRepeaters(ActiveLevel)    'Level 0 sensor binons repeat if they have the same intensity
                
                'Now need to combine overlapping shape patterns or adjacent repeating patterns.
                ' If it is a pattern (overlapable) and overlaps with the next one and the next one is a pattern
                'then create the pattern of the two patterns. They are put into the next level as a pattern.
        Call CombineBinons(ActiveLevel)     'Combine ActiveBinonTree Shape binons found in source level ActiveLevel
        ActiveLevel = ActiveLevel + 1       'Process the next level up in ActiveBinonTree()
      Wend

End Sub         'End of RecognizeDigit()

Public Sub DoCenterSurround(SensorValues(), SensorLevel)

    'At the SensorLevel = 0 of ActiveBinonTree() we need to go sequentially through positions
    'calculating the On center and Off center values using the IDL ratios to the left and right of the sensor position.
    'Which ever value is greater than zero, use it for that sensor position.
    'Then go through these values looking for the edges - transitions from on to off or vice versa.
    'Halfway between the on and off is the location of the edge.
    'CountRepeaters will combine adjacent same pixels into a range with a size/width/quantity.
    
'  Public ActiveBinonTree&(0 To 25, 1 To 64, 1 To 3) 'The tree of the Shape Binons found in the image.
'                                        'Dynamically recreated as each image is processed/experienced/perceived
'                                        '1st index = which level is being processed
'                                        '2nd index = The active entry - corresponds to a sensor position
'                                        '3rd index = Activation property values
'                '---- ActiveBinonTree() types of property values
'  Public PatternSize    '1, The size of the pattern - measured in number of sensors involved at level 0
'  Public Quantity       '2, Numerosity, quantity of times this pattern repeats - zero if entry is to be skipped
'  Public TheBinon       '3, The binon / pattern experienced - if any - ShapeBin

  Dim SensorPosition
  Dim PreviousIntensity, Intensity, NextIntensity
  Dim LeftRatio, RightRatio
  Dim OnValue, OffValue
  Dim LastOnValue, LastOffValue
  Dim LastOnPosition, LastOffPosition
  Dim EdgePosition, HaveEdge As Boolean
  Dim UpEdge, DownEdge, NoEdge, TheEdge
  Dim BlackBinon&, WhiteBinon&
  
                '-----  These three values are kept in ActiveBinonTree(SensorLevel, SensorPosition, Quantity) to indicate
                '       what type of edge is at this position
    UpEdge = 1                'Indicates an edge where the intensity increases
    DownEdge = -1             'Indicates an edge where the intensity decreases
    NoEdge = 0                'Indicates there is no edge at this SensorPosition
    
                'The BlackBinon is put in ActiveBinonTree() between an UpEdge and the next DownEdge.
        'IdBinons(BinLevel, BinType, Bin1&, Bin2&, IDLRatio, BinonID&)      'Create new or find existing binon
    Call IdBinons(ActiveLvl, ShapeBin, NullBinon, NullBinon, 17, BlackBinon)
    
                'The ActiveBinonTree() contains WhiteBinons between a DownEdge and the next UpEdge
    Call IdBinons(ActiveLvl, ShapeBin, NullBinon, NullBinon, 1, WhiteBinon)
                                      
    PreviousIntensity = 1                                     'Intensity at sensor #0 off the left end of array = white=1
    LastOnValue = 0
    LastOffValue = 0
    
    For SensorPosition = 1 To NumberOfSensors                 'For all sensor intensity values
      
      HaveEdge = False                                        'assume we do not have an edge
      Intensity = SensorValues(SensorPosition)                'get the intensity at the sensor position
      If SensorPosition < NumberOfSensors Then
        NextIntensity = SensorValues(SensorPosition + 1)
      Else
        NextIntensity = 1                                     'Intensity at sensor #65 off the right end of array = white=1
      End If
          
      LeftRatio = IDLof(Intensity, PreviousIntensity)         'Log of Intensity/PreviousIntensity
      RightRatio = IDLof(Intensity, NextIntensity)            'Log of Intensity/NextIntensity
      OnValue = LeftRatio + RightRatio
      If OnValue < 0 Then OnValue = 0
      
      LeftRatio = IDLof(PreviousIntensity, Intensity)         'Log of PreviousIntensity/Intensity
      RightRatio = IDLof(NextIntensity, Intensity)            'Log of NextIntensity/Intensity
      OffValue = LeftRatio + RightRatio
      If OffValue < 0 Then OffValue = 0
          
      If OffValue > 0 Then                                    'If have an Off value
        If LastOnValue > 0 Then                               'and it follows an On value then have a decreasing edge
                'The EdgePosition is half way between this off position and the last on position
                'This edge position will not always be at the SensorPosition
          EdgePosition = (SensorPosition + 0.1 - LastOnPosition) / 2 + LastOnPosition     'Round up if X.5
                'Select the UpDown Binon for the slope, PosBinon, NegBinon at this edge position
                'The source Intensities are in ShapBin entries in ActiveBinonTree() so need to update - not add new entries
          ActiveBinonTree(SensorLevel, EdgePosition, Quantity) = DownEdge   'not always at the SensorPosition
          HaveEdge = True
        End If
        LastOffValue = OffValue                               'have another OffValue so keep the latest
        LastOffPosition = SensorPosition
        LastOnValue = 0
        
      ElseIf OnValue > 0 Then                                 'else If have an On Value
        If LastOffValue > 0 Then                              'and it follows an Off value then have an increasing edge
                'The EdgePosition is half way between this on position and the last off position
                'This edge position will not always be at the SensorPosition
          EdgePosition = (SensorPosition + 0.1 - LastOffPosition) / 2 + LastOffPosition   'Round up if X.5
                'Select the UpDown Binon for the slope, PosBinon, NegBinon at this edge position
                'The source Intensities are in ShapBin entries in ActiveBinonTree() so need to update - not add new entries
          ActiveBinonTree(SensorLevel, EdgePosition, Quantity) = UpEdge   'not always at the SensorPosition
          HaveEdge = True
        End If
        LastOnValue = OnValue                                 'have another OnValue so keep it as the latest
        LastOnPosition = SensorPosition
        LastOffValue = 0
      Else                                                    'both are zero so go to the next position
      End If
      
      If Not HaveEdge Then                                    'If no edge found then HaveEdge is false
        ActiveBinonTree(SensorLevel, SensorPosition, Quantity) = NoEdge
      Else
      End If
        
      PreviousIntensity = Intensity                           'keep this intensity for comparison purposes
      
    Next SensorPosition
    
                'At this point all the shape binons are filled in with either UpEdge, DownEdge or NoEdge
                'Now replace the UpEdge and NoEdge upto the next DownEdge with FlatBinon and
                'replace the DownEdge with a NullBinon
    HaveEdge = False                                          'set true when first find an UpEdge, False for an DownEdge
    For SensorPosition = 1 To NumberOfSensors                 'For all sensor intensity values
          
      TheEdge = ActiveBinonTree(SensorLevel, SensorPosition, Quantity)
      If TheEdge = UpEdge Then
        HaveEdge = True                                       'It will stay true until the next DownEdge
      ElseIf TheEdge = DownEdge Then
        HaveEdge = False
      End If
      
      If HaveEdge Then
            'SaveActiveBinon(ActiveLvl, Entry, SizeOfPattern, Quanty, BinonID&)
                                                              'save the black binon in level 0 of ActiveBinonTree()
        Call SaveActiveBinon(SensorLevel, SensorPosition, 1, 1, BlackBinon)

      Else          'Else one of the sensors on either side of this edge have changed in intensity
                                                              'save the white binon in level 0 of ActiveBinonTree()
        Call SaveActiveBinon(SensorLevel, SensorPosition, 1, 1, WhiteBinon)

      End If

    Next SensorPosition

End Sub               'End of DoCenterSurround()

Public Sub SaveActiveBinon(ActiveLvl, Entry, SizeOfPattern, Quanty, BinonID&)

      'Places the values for the binon in the ActiveBinonTree(ActiveLvl) at the Entry position
      'The place (entry) is the same as a sensor position in the array of sensors
      
'  Public ActiveBinonTree&(0 To 25, 1 To 500, 1 To 5) 'The tree of the Shape Binons found in the frame
                                        'dynamically recreated as each frame is processed/experienced/perceived
                                        '1st index = which level is being processed
                                        '2nd index = The active entry - corresponds to a sensor position
                                        '3rd index = Activation property values

                '1, Size of the pattern - measured in number of sensors involved at level-0
      ActiveBinonTree(ActiveLvl, Entry, PatternSize) = SizeOfPattern
                '2, Numerosity, quantity of times this pattern repeats - zero if it is to be skipped
      ActiveBinonTree(ActiveLvl, Entry, Quantity) = Quanty
                '3, The binon experienced - a ShapeBin
      ActiveBinonTree(ActiveLvl, Entry, TheBinon) = BinonID
      
      If BinonID <> NullBinon Then
        Call AddToAllFoundBinons(BinonID)       'add it to the AllFound list (if not already on it)
      End If
    
End Sub           'End of SaveActiveBinon()

Public Sub CountRepeaters(ActiveLvl)       'Repeater counting algorithm

        'Processes ShapeBin binons. Any that are adjacent, repeat and are the same size are counted to set the quantity
        'This results in changes to the patterns at the ActiveLvl, it does not generate any combinations at the next level up

  Dim ARepeater As Boolean
  Dim Place1, Qty1, Bin1&, SizeOfPattern1, LogSize1
  Dim LastPlace, SizeOfLastPattern
  Dim WantPlace2, Qty2, Bin2&, SizeOfPattern2, LogSize2
  Dim Place3
      
      Place1 = 1                                                    'For the 64 entries in ActiveBinonTree()
      While Place1 < NumberOfSensors
        Qty1 = ActiveBinonTree(ActiveLvl, Place1, Quantity)         'The count of this pattern
        If Qty1 = 1 Then                                            'It will be zero if the entry is to be skipped - empty
          SizeOfPattern1 = ActiveBinonTree(ActiveLvl, Place1, PatternSize)    'Need the sizes of this pattern
          LogSize1 = Log_Based_on_JND(SizeOfPattern1)               'The log size for first pattern x 100
          LastPlace = Place1                                        'start with the 1st occurrence of the pattern
                                                                    'will be set to the entry of the last repeater found
          SizeOfLastPattern = SizeOfPattern1                        'adjusted for each repeat pattern that is found
          
          Do                                                        'Do until a repeater not found or off the end
            ARepeater = False                                       'Will be set if find a repeater, then look for next repeat
                          'To be a repeater it must be adjacent to the previous pattern at LastPlace
            WantPlace2 = LastPlace + SizeOfLastPattern              'possible position of next repeater
            If WantPlace2 <= NumberOfSensors Then                   'Provided not off the end of the list
              Qty2 = ActiveBinonTree(ActiveLvl, WantPlace2, Quantity) 'The count of the pattern adjacent to the last one
              If Qty2 > 0 Then                                      'It will be zero if it is empty - no repeater
              
                          'To be a repeater it must be the same binon
                Bin1 = ActiveBinonTree(ActiveLvl, Place1, TheBinon)     'The binon at this location
                Bin2 = ActiveBinonTree(ActiveLvl, WantPlace2, TheBinon) 'The binon that is next to it
                If Bin1 = Bin2 Then                                 'If it is the same pattern
              
                          'To be a repeater it must have a size within a JND of the first one of the repeaters
                  SizeOfPattern2 = ActiveBinonTree(ActiveLvl, WantPlace2, PatternSize)    'Need the sizes of the next pattern
                  LogSize2 = Log_Based_on_JND(SizeOfPattern2)       'The log size for the pattern x 100 for comparison
                  If Fix((LogSize1 - LogSize2) / 100) = 0 Then      'If the difference in the sizes is not noticeable then
                    LastPlace = WantPlace2                          'position of this repeater
                    SizeOfLastPattern = SizeOfPattern2              'to calculate location of next possible repeater pattern
                    Qty1 = Qty1 + 1                                 'count it
                    ActiveBinonTree(ActiveLvl, Place1, Quantity) = Qty1   'count of repetitions
                    SizeOfPattern1 = SizeOfPattern1 + SizeOfPattern2  'new combined size of repeating pattern
                    ActiveBinonTree(ActiveLvl, Place1, PatternSize) = SizeOfPattern1
                    ARepeater = True
                  Else                                  'else SizeOfPattern1 <> SizeOfPattern2, ARepeater is False
                  End If
                Else                                    'else not the same pattern, ARepeater is False
                End If
              Else                                      'else no binon at WantPlace2, ARepeater is False
              End If
            Else                                        'else off the end of the list of entries, ARepeater is False
            End If
          Loop Until Not ARepeater                      'Do it again until not a repeater or off the end of the list
          
          If Qty1 > 1 Then                                          'If a repeater was found then
                                          'From the place after the first pattern to the end of the last repeating pattern
                                          'for all the inbetween binons, including the repeating ones
            For Place3 = Place1 + 1 To LastPlace + SizeOfLastPattern - 1
              Call ClearActiveBinonEntry(ActiveLvl, Place3)         'clear the entry in the ActiveBinonTree()
            Next Place3
            Place1 = WantPlace2                                     'the next entry after the repeater, if any
          Else                                          'else no repeater found so just progress to the next pattern
            Place1 = Place1 + 1
          End If
        Else                                            'else quantity at Place1 = 0 so skip it
          Place1 = Place1 + 1                                       'go to the next place
        End If                                          'End If Qty1 > 0
      Wend
      
End Sub             'End of CountRepeaters()

Public Sub CombineBinons(ActiveLvl)         'Combine ActiveBinonTree of ShapeBin binons

    'Called from RecognizeDigit to process the ActiveLevel of ActiveBinonTree() and put the combined Shape binons
    'in level NewActiveLevel = ActiveLevel+1 of ActiveBinonTree(), Starts with lowest level of ActiveLvl = 0 = sensor binons
    
          'Combines patterns into bigger patterns. The two source binons are the trigger source binon1 and goal source binon2
          'Two patterns can only overlap if they are both quantity=1 and they share a common subpattern.
          'A pattern can be combined with an adjacent pattern if one or both of them are repeaters, quantity > 1.
          'When two source pattern binons are combined adjacently the ratio of the target binon is based on
          'the repeat quantities of the two source binons.
          

    Dim NewActiveLevel
    Dim Place1, SizeOfPattern1, Bin1&, Intrst1, Qty1, Bin12&
    Dim Place2, SizeOfPattern2, Bin2&, Intrst2, Qty2, Bin21&, AdjacentPlace2
    Dim NewPatternSize
    Dim LogQty1, LogQty2, QuantityRatio
    Dim CanCombine As Boolean, CombinedBinon&
      
    NewActiveLevel = ActiveLvl + 1                                        'The next level higher in the Activation Tree
    HighestActiveLevel = NewActiveLevel                                   'Keep track of the highest active level produced
    
    For Place1 = 1 To NumberOfSensors - 1
      
      Qty1 = ActiveBinonTree(ActiveLvl, Place1, Quantity)                 'The repeat count of this pattern at Place1
      If Qty1 > 0 Then                                                    'zero if it is to be skipped - empty entry
                        'The position of the combined target pattern will be the same as its left source binon
        SizeOfPattern1 = ActiveBinonTree(ActiveLvl, Place1, PatternSize)  'Need the total size of the pattern (sensors spanned)
        CanCombine = False                                                'Assume we can not combine Place1 and Place2 binons
                    
              'At level 0 adjacent sensors with the same reading result in a repeating group with a width equal to
              'the number of sensors and these widths are adjacent to each other. A repeating pattern is always
              'combined adjacently with the pattern beside it.
              'Check to see if there is an adjacent pattern - if so combine adjacent patterns and proceed to next pattern
      '--- ADJACENT ---                                 'Repeating patterns combine adjacently to other patterns
        AdjacentPlace2 = Place1 + SizeOfPattern1        'where the next adjacent pattern would be
        If AdjacentPlace2 <= NumberOfSensors Then                         'Provided not off the end of the list
          Qty2 = ActiveBinonTree(ActiveLvl, AdjacentPlace2, Quantity)     'The Qty2 will be zero if no adjacent pattern
          If Qty2 > 0 Then                                                'If there is an adjacent pattern then
            If ActiveLvl = 0 Or Qty1 > 1 Or Qty2 > 1 Then 'If one of the two patterns is a repeater it must be combined adjacently
          
              Bin1 = ActiveBinonTree(ActiveLvl, Place1, TheBinon)         'The binon at this location
              Bin2 = ActiveBinonTree(ActiveLvl, AdjacentPlace2, TheBinon) 'The binon next to it at Place2
              SizeOfPattern2 = ActiveBinonTree(ActiveLvl, AdjacentPlace2, PatternSize)   'NewPatternSize is the total new size.
              NewPatternSize = SizeOfPattern1 + SizeOfPattern2
              Intrst1 = InterestIn(Bin1)                                  'Interest in source binon
              Intrst2 = InterestIn(Bin2)                                  'Its at sensor level so should be familiar
              If Intrst1 = FAMILIAR And Intrst2 = FAMILIAR Then           'both source binons must be familiar
                        '   The IDL ratio of the target binon is based on the relative (size) quantity of the source binons.
                QuantityRatio = IDLof(Qty1, Qty2)
              
                CanCombine = True                                         'Combine this one and next and place in level above.
              Else                                        'else adjacent patterns are not both familiar, can not combine
              End If                                      'End if the two adjacent patterns are both familiar
            Else                                          'Else adjacent but both quantity=1, check for overlappers instead
            End If                                        'End if either of the two adjacent patterns are repeaters
          Else                                            'Else there is no adjacent pattern
          End If                                          'End if there is a pattern at the adjacent place
        Else                                              'else off the end of the list of entries
        End If
          
      '--- OVERLAPPING ---                          'Two patterns with a common subpattern
        If Not CanCombine Then                      'If didn't find an adjacent pattern then is there an overlapping one
              'If Place1 binon Qty=1 then check if the next one is quantity=1 and overlaps - then combine them
          If Qty1 = 1 Then                          'If the Place1 pattern is not a repeater (quantity=1) then
                          'Search to the right of the Place1 pattern up to the end of Place1 positions
                          'which is the position before the possible AdjacentPlace2
                          'for the first pattern found and see if it overlaps with the Place1 pattern
            For Place2 = Place1 + 1 To AdjacentPlace2 - 1                 'go thru places to the right up to end of Bin1
              Qty2 = ActiveBinonTree(ActiveLvl, Place2, Quantity)         'The repeat quantity of the first pattern found
                                                                          'Qty2=0 if there is no pattern at this place
              If Qty2 > 0 Then
                If Qty2 = 1 Then                          'If overlapping then the goal source binons must not repeat
                  Bin1 = ActiveBinonTree(ActiveLvl, Place1, TheBinon)     'The binon at this location
                  Bin12 = Binons(Bin1, GoalSource)                        'The 2nd (goal) source binon of Bin1
                  Intrst1 = InterestIn(Bin1)                              'Interest in source binon
                  Bin2 = ActiveBinonTree(ActiveLvl, Place2, TheBinon)     'The binon at this position - next to it or further away
                  Bin21 = Binons(Bin2, TriggerSource)                     'The 1st (trigger) source binon of Bin2
                  Intrst2 = InterestIn(Bin2)              'Bin12 = the second (goal) source pattern of binon 1
                                                          'If they are both familiar and have a common subpattern
                  If Intrst1 = FAMILIAR And Intrst2 = FAMILIAR And Bin12 = Bin21 Then
                    SizeOfPattern2 = ActiveBinonTree(ActiveLvl, Place2, PatternSize)  'NewPatternSize is the total new size.
                    NewPatternSize = Place2 - Place1 + SizeOfPattern2       '
                          'When combining overlapping patterns the IDL Ratio of the resulting pattern is the ratio
                          'of Qty1 to Qty2 which is 1/1 and the log of this is zero
                    QuantityRatio = 0                     'the difference in the Logs = ratio of quantities = IDL
                        
                    CanCombine = True
                  Else                                    'Else the two overlapping binons are not both familiar
                                                          'or they do not share a common subpattern and can not be combined
                  End If                                  'End if the two are familiar and combinable
                Else                                      'Else Qty2 must be > 1 so adjacents have been checked
                End If                                    'End if Place2 Qty2 = 1
                                                          'Terminate the loop - only the first one found is tested
                Place2 = NumberOfSensors                  'Whether combinable or not combinable - search no further
              Else                                        'Else Qty2 = 0 so check at next Place2
              End If
            Next Place2
          Else                                            'Else Qty1 must be > 1 so can only combine adjacently
          End If                                          'End if Place1 Binon is not a repeater
        End If                                            'End if there was no adjacent combination
                                    
        If CanCombine Then                                'Create target shape binon as a pattern
        
              'IdBinons(BinLevel, BinType, Bin1&, Bin2&, IDLRatio, BinonID&)  'Create new or find existing binon
          Call IdBinons(NewActiveLevel, ShapeBin, Bin1, Bin2, QuantityRatio, CombinedBinon)
          If OverLimit Then Exit Sub
          
              'SaveActiveBinon(ActiveLvl, Entry, SizeOfPattern, Quanty, BinonID&)
          Call SaveActiveBinon(NewActiveLevel, Place1, NewPatternSize, 1, CombinedBinon)
          
          CountCombined = CountCombined + 1               'increment count of NewActiveLevel entries
        Else                                              'Else the source patterns are not combinable
        End If
      Else                                                'Else if quantity of Place1 entry is zero then skip it
      End If                                              'End If quantity of binons at this Place1 position is > 0
    Next Place1

End Sub           'End of CombineBinons()

Public Sub ClearActivationTree()

    Dim ActiveLvl, Entry

        For ActiveLvl = 0 To HighestActiveLevel         'Clear out the active binon history for the last image processed
          
          For Entry = 1 To NumberOfSensors
            Call ClearActiveBinonEntry(ActiveLvl, Entry)
          Next Entry
            
        Next ActiveLvl
        
        HighestActiveLevel = 0                          'so far no active binons
      
End Sub           'End of ClearActivationTree()

Public Sub ClearActiveBinonEntry(ActiveLvl, Entry)

          ActiveBinonTree(ActiveLvl, Entry, Quantity) = 0
          ActiveBinonTree(ActiveLvl, Entry, PatternSize) = 0
          ActiveBinonTree(ActiveLvl, Entry, TheBinon) = NullBinon

End Sub           'End of ClearActiveBinonEntry()

Public Sub Reset_Adaptron()
          'Resets all changing information to starting places.
          'Must be called before setting up any sensors.
          'Called from Finish_Load_Process
          
  Dim I&
          
  Call MainDisplay.ClearMemory_Click        'This will zero / initialize all values and set JND to 20%
  
  Call Reset_Accuracy
                                            'Initialize AssociationLinks to all free
  For I = 1 To MaxBinons - 1
    AssociationLinks(I, NextLink) = I + 1   'Each one points to the next one, The last one points to nulllink = end of list
  Next I
  FreeLink = 1                              'First free AssociationLinks()
  
End Sub           'End of Reset-Adaptron()

Public Sub Reset_Accuracy()     'Reset numbers that keep track of accuracy of predictions

  NextEntryIn100 = 1
  NumberOfEntriesInLast_100 = 0
  
  MainDisplay.PercentCorrect.Text = "0 %"
  For I = 0 To 9
    MainDisplay.DigitCorrect(I).Text = "0 %"
    MainDisplay.CountDigits(I).Text = "of 0"
    CorrectDigit(I) = 0
    DigitCount(I) = 0           'Number of times this digit experienced - used in ReportPrediction
  Next I

  For I = 0 To 12
    LvlCorrect(I) = 0           'Number of correct predictions at this level
    LvlPredictions(I) = 0       'Number of predictions at this level
    MainDisplay.Corrects(I).Text = "0 %"
    MainDisplay.Predictions(I).Text = "of 0"
  Next I
  MainDisplay.TotalPredictions.Text = "of 0"
  MainDisplay.NoPrediction.Text = "0"
  NoPredictionMade = 0
  MainDisplay.CorrectTotal.Text = "0 %"
  
End Sub         'End of Reset-Accuracy()

Public Sub Fill_In_Pixel_Display_Of_Digit(ArrayEntry)

  Dim I, Valu&, Pixel_Value
  
    If ArrayEntry < Number_of_Digits Then
    
      For I = 0 To NumberOfSensors - 1
        Pixel_Value = DigitPixels(ArrayEntry, I Mod 8, I \ 8)       'Values of 0 = white or 16 =  black
        
        If BitmapsOrImages$ = Bitmaps$ Then

          If Pixel_Value > 7 Then
            Pixel_Value = 16                    'Values 8 to 16 are black
          Else
            Pixel_Value = 0                     'Values 0 to 7 are white
          End If
          
        Else                                    'If BitmapsOrImages$ = Images$ Then no change in pixel values
        End If
  
        Valu& = (16 - Pixel_Value) * 16         'converts 0 to 256 for white and 16 to 0 for black
        If Valu& = 256 Then Valu& = 255
        MainDisplay.Pixel(I).BackColor = Valu& + Valu& * 256 + Valu& * 65536   '&HFFFFFF = white, &H000000 = Black
          
        MainDisplay.TheDigit.Text = ActualDigit(ArrayEntry)
        MainDisplay.Entry.Text = Trim(Str$(ArrayEntry)) & "  of" & Str$(Number_of_Digits - 1)
      Next I
      
    End If
    
End Sub             'End of Fill-In-Pixel-Display-Of-Digit()

Public Sub Get8x8Digits()         'Load the 8x8 image pixel values and digit labels

  Dim FirstPosition, SecondPosition, NumberText$, X, Y, RealLine$, Pixel_Value
  
  Open "Hand Drawn Digits 8x8.txt" For Input As #1
  
  Number_of_Digits = 1            'index into the the values of the handwritten digits, 0 to 9
  
  While Not EOF(1)
  
    Line Input #1, RealLine$      'digits are seperated by commas, lines by CR LF
  
      FirstPosition = 1
      SecondPosition = 1
      X = 0                       'horizontal position of pixel 0 to 7
      Y = 0                       'vertical position of pixel 0 to 7
      
      While SecondPosition > 0
        SecondPosition = InStr(FirstPosition, RealLine$, ",")     'find the next comma
        If SecondPosition > 0 Then
          NumberText$ = Mid$(RealLine$, FirstPosition, SecondPosition - FirstPosition)    'get the string of text
          Pixel_Value = Val(NumberText$)                          'Values range between 0 to 16
          
          DigitPixels(Number_of_Digits, X, Y) = Pixel_Value
            
          X = X + 1
          If X = 8 Then
            Y = Y + 1
            X = 0
          End If
      
          FirstPosition = SecondPosition + 1                      'reposition pointers
        Else                      'SecondPosition=0 so FirstPosition to end is the DigitVal
          NumberText$ = Mid$(RealLine$, FirstPosition, Len(RealLine$))
          ActualDigit(Number_of_Digits) = Val(NumberText$)
        End If
      Wend
      Number_of_Digits = Number_of_Digits + 1
  Wend
  
  Close #1

End Sub             'End of Get8x8Digits()

Public Sub MakePrediction()

      'Called from ProcessNextDigit_Click() after RecognizeDigit() has finished
      'After all the Active binon tree entries have been produced we might have a prediction value for the image.
      
  Dim ActualDigitLabel, I, HighestLevel
  Dim HighestCount, DigitFound
  
'  Public FoundBinons&(0 To 100, 3) 'Up to 100 binons found with (0, 0) = count of entries
                '---- FoundBinons() properties - its 2nd index
'  Public TheFoundBinon            ' 0 = Found Binon ID
'  Public TheLabelBinon            ' 1 = Label Binon ID
'  Public TheBinonLevel            ' 2 = Binon level
                                  
    PredictedDigitLabel = -1            'The digit prediction = -1 if no prediction made
    
    If FoundBinons(0, 0) > 0 Then       'If have at least one prediction then
    
                  'Use the most frequently occurring label of the FoundBinons
          For I = 0 To 9                'Clear out the highest levels at which the 10 digit values were found
            HighestFindLevel(I) = 0
          Next I
          
'  Public FoundBinons&(0 To 100, 3) 'Up to 100 binons found with (0, 0) = count of entries
                '---- FoundBinons() properties - its 2nd index
'  Public TheFoundBinon            ' 0 = Found Binon ID
'  Public TheLabelBinon            ' 1 = Label Binon ID
'  Public TheBinonLevel            ' 2 = Binon level
                                  
                  'Use the most frequently occurring one
          For I = 1 To FoundBinons(0, 0)          'For all the predictive shape binons found
            DigitFound = Binons(FoundBinons(I, TheLabelBinon), Ratio)    'get the digit value of the Label Binon
                                                  'increment the number of finds for this digit label
            FindCountForDigit(DigitFound) = FindCountForDigit(DigitFound) + 1
            If FoundBinons(I, TheBinonLevel) > HighestFindLevel(DigitFound) Then
              HighestFindLevel(DigitFound) = FoundBinons(I, TheBinonLevel) 'The highest ActiveBinonTree level digit found
            End If
          Next I
          
          HighestCount = 0                        'Find the highest quantity of occurrences
                    'Note: If two or more digits have the same find count then the higher digit is selected
                    'This will tend to select 9 more frequently
          For I = 9 To 0 Step -1
            If FindCountForDigit(I) > HighestCount Then
              PredictedDigitLabel = I             'The digit with the highest FindCountForDigit()
              HighestLevel = HighestFindLevel(I)  'The highest level it was found at
              HighestCount = FindCountForDigit(I) 'new highest number of occurrences
            End If
          Next I
      
'   ---------------------  Update the Accuracy information --------------

      ActualDigitLabel = Val(ActualDigit(Stimulus_Number))
      DigitCount(ActualDigitLabel) = DigitCount(ActualDigitLabel) + 1       'Number of times this digit experienced - used in ReportPrediction
      If HighestLevel > 11 Then HighestLevel = 12
      LvlPredictions(HighestLevel) = LvlPredictions(HighestLevel) + 1
      If PredictedDigitLabel = ActualDigitLabel Then
        CorrectInLast50 = CorrectInLast50 + 1                 'count how many correctly predicted in last 50
        Call UpdateLast100Accuracy(True)
        CorrectDigit(ActualDigitLabel) = CorrectDigit(ActualDigitLabel) + 1
        LvlCorrect(HighestLevel) = LvlCorrect(HighestLevel) + 1
      Else        'incorrect prediction
        Call UpdateLast100Accuracy(False)
      End If
      
    Else      'Else FoundBinons(0,0) = 0 and PredictedDigitLabel=-1 - no prediction made counts as incorrect prediction
      NoPredictionMade = NoPredictionMade + 1
    End If
    
End Sub       'End of MakePrediction()

Public Sub AddToAllFoundBinons(BinonFound&)

        'Called by SaveActiveBinon()
        
    Dim Entry, AlreadyFound As Boolean
  
      AlreadyFound = False
      For Entry = 1 To AllFoundBinons(0)              'Go through the AllBinons found list
        If AllFoundBinons(Entry) = BinonFound Then    'to see if BinonFound is already on the list
          AlreadyFound = True
          Entry = AllFoundBinons(0)                   'terminate loop
        End If
      Next Entry
      
      If Not AlreadyFound Then                        'add it to the list if it is not already on it.
        AllFoundBinons(0) = AllFoundBinons(0) + 1
        AllFoundBinons(AllFoundBinons(0)) = BinonFound
        Call AddPrediction(BinonFound)                'add any associated label binons to FoundBinons() list
      End If
      
End Sub             'End of AddToAllFoundBinons()

Public Sub AddPrediction(FoundBinon&)

      'Called from AddToAllFoundBinons()
      
      'As active binons are recognized they are checked here to see if they have any associated label binons.
      'If they have more than one associated label binon then the assocated label is ambiguous and is not used for prediction
      'If they have just one associated label binon then the pattern uniquely appears in the one digit image,
      'the digit is predicted.
      'The prediction information is added to the FoundBinons() list. This information gets used
      'by MakePrediction after all the digit pixels have been processed.
      
      Dim LabelBinon&, AssociationLink&, AssociationCount, FoundListEntry
      
'  Public FoundBinons&(0 To 100, 3) 'Up to 100 binons found with (0, 0) = count of entries
                '---- FoundBinons() properties - its 2nd index
'  Public TheFoundBinon            ' 0 = Found Binon ID
'  Public TheLabelBinon            ' 1 = Label Binon ID
'  Public TheBinonLevel            ' 2 = Binon level
                                  
          LabelBinon = NullBinon                    'Assume no associated label binon
               
          GoSub CheckTargets                        'Find out if it is unique and set the LabelBinon to its LabelBinon
          
          If LabelBinon <> NullBinon Then           'we have found a single Associated label binon for the Shape binon
            FoundListEntry = FoundBinons(0, 0) + 1            'next available FoundBinons entry
            If FoundListEntry > 100 Then FoundListEntry = 1   'If have 100 entries start replacing previous ones
            FoundBinons(0, 0) = FoundListEntry                'The number of FoundBinons() entries
            FoundBinons(FoundListEntry, TheFoundBinon) = FoundBinon                         'The binon found
            FoundBinons(FoundListEntry, TheLabelBinon) = LabelBinon                         'The associated label binon
            FoundBinons(FoundListEntry, TheBinonLevel) = Binons(FoundBinon, BinonLevel)     'the binon level
          End If
Exit Sub

CheckTargets:               'The shape pattern is ambiguous if it has more than one associated label binon
                            'The shape pattern uniquely predicts its digit label if there is only one associated label binon
                            'The shape pattern is unclassified if it has no associated label binons
        AssociationLink = Binons(FoundBinon, AssociationList)
        If AssociationLink <> NullLink Then     'If it has an AssociationLinks() pointer
                                                'And if it only has this one AssociationLinks entry then it is unique
          If AssociationLinks(AssociationLink, NextLink) = NullLink Then    'It does not point to a second one in linked list
                                  
            LabelBinon = Binons(AssociationLinks(AssociationLink, AssocBinonID), GoalSource) 'The associated label Binon
            
          End If
        End If
  Return

End Sub             'End of AddPrediction()

Public Sub Associate_Actual_Label_With_Found_Shapes()

      'Called from RecognizeDigit() after it has made a prediction for the image
      
      'Now that we know what it predicted the digit to be (PredictedDigitLabel from the FoundBinons() list)
      'we can update the shape binons in the AllFoundBinons() list to associate with the actual digit label
      'Go through the AllFoundBinons() list and associate the label binon with all the ShapeBin binons found in the image

  Dim Entry, FoundBinID&, ActualDigitLabel
  Dim AssociationBinon&, AssociationLink&, PreviousLink&, LabelBinon&
  Dim AddIt As Boolean
  
    ActualDigitLabel = Val(ActualDigit(Stimulus_Number))
    
                        'Find / create the symbolic label binon for this actual digit
        'IdBinons(BinLevel, BinType, Bin1&, Bin2&, IDLRatio, BinonID&)  'Create new or find existing binon
    Call IdBinons(0, LabelBin, NullBinon, NullBinon, ActualDigitLabel, LabelBinon)
    
    For Entry = 1 To AllFoundBinons(0)                'For all found binons
    
      FoundBinID = AllFoundBinons(Entry)              'Obtain the binon ID
                        'Create an association binon between the Found binon and the label binon,
          'IdBinons(BinLevel, BinType, Bin1&, Bin2&, IDLRatio, BinonID&)  'Create new or find existing binon
      Call IdBinons(1, AssociationBin, FoundBinID, LabelBinon, 1, AssociationBinon)
      If OverLimit Then Exit Sub
      
                        'Add the association binon to a linked list of them for this found shape binon
            'See if the binon is already associated with this label binon
      AssociationLink = Binons(FoundBinID, AssociationList)   'get the 1st one on its linked list of AssociationBinons associated to label binons
      PreviousLink = AssociationLink              'will not be used if label binon already on list
      AddIt = True                                'Assume will add the AssociationBinon to the AssociationList of the shape
      While AssociationLink <> NullLink And AddIt
        If AssociationLinks(AssociationLink, AssocBinonID) = AssociationBinon Then  'if this is the same AssociationBinon associated to the label binon then
          AddIt = False                           'terminate search
        Else
          AssociationLink = AssociationLinks(AssociationLink, NextLink)  'go to the next associated binon in the link list
        End If
      Wend

      If AddIt Then                               'If still true then add the association to the linked list
        AssociationLink = FreeLink                'use the next free AssociationLinks() entry
        FreeLink = AssociationLinks(FreeLink, NextLink)   'Move the free pointer to the next free one
        If FreeLink = NullLink Then Stop                  'run out of free AssociationLinks() - major problem
            'Put the new AssociationBinon on the AssociationList of the AssociationBinons
        AssociationLinks(AssociationLink, NextLink) = PreviousLink          'put it at the start of the linked list
        AssociationLinks(AssociationLink, AssocBinonID) = AssociationBinon  'Point the list entry to the AssociationBinon associated to the label binon
        Binons(FoundBinID, AssociationList) = AssociationLink   'Put it in as the AssociationList first linked list item of the Binon
      End If
    Next Entry                                      'next binon in the All Found list

End Sub           'End Associate-Actual-Label-With-Found-Shapes()

Public Sub UpdateLast100Accuracy(CorrectPrediction As Boolean)

    'Called from MakePrediction()
    
    '  Public IsACorrectPrediction(1 To 100) As Boolean  'true if it was correctly predicted
    '  Public NextEntryIn100                'location in IsACorrectPrediction() to put the next one
    '  Public NumberOfEntriesInLast_100     'number of entries in IsACorrectPrediction()

    IsACorrectPrediction(NextEntryIn100) = CorrectPrediction                'Flag the entry as correctly predicted or not
 
    NextEntryIn100 = NextEntryIn100 + 1
    If NextEntryIn100 > 100 Then NextEntryIn100 = 1                         'Loop back to 1st entry when 100 reached
 
    NumberOfEntriesInLast_100 = NumberOfEntriesInLast_100 + 1               'May have fewer than 100 entries to start with
    If NumberOfEntriesInLast_100 > 100 Then NumberOfEntriesInLast_100 = 100 'Never more than 100 entries
 
End Sub             'End of UpdateLast100Accuracy()

Public Function IDLof(FirstValue, SecondValue)

  Dim LogOfFirst, LogOfSecond
  
                                                          'The IDL ratio is based on the relative values
      LogOfFirst = Log_Based_on_JND(FirstValue)           'The log of the 1st value * 100
      LogOfSecond = Log_Based_on_JND(SecondValue)         'The log of the 2nd value * 100
      IDLof = Fix((LogOfFirst - LogOfSecond) / 100)       'The integer difference in the Logs = ratio of 1st/2nd

End Function        'End of IDLof()

Public Function Log_Based_on_JND(Valu)      'Calculate the log of the Valu using a base of LogBase and multiply by 100

    Log_Based_on_JND = 100 * Log(Valu) / Log(LogBase#)
  
End Function          'End of Log_Based_on_JND()

